from contextlib import AsyncExitStack

from anyio import create_task_group


class StartStopContextManager:
    def __init__(self, service):
        self._service = service

    async def __aenter__(self):
        async with AsyncExitStack() as exit_stack:
            self._task_group = await exit_stack.enter_async_context(create_task_group())
            await self._task_group.start(self._service.start)
            self._exit_stack = exit_stack.pop_all()
        await self._service.started.wait()
        return self._service

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        self._task_group.start_soon(self._service.stop)
        return await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)



import gc

import pytest


def finalize():
    gc.collect()
    # looks like PyPy needs a second one:
    gc.collect()


@pytest.fixture(scope="function", autouse=True)
def collect_gc(request):
    request.addfinalizer(finalize)


@pytest.fixture(params=("context_manager", "start_stop"))
def service_api(request):
    return request.param



import json
from copy import deepcopy
from uuid import uuid4

import pytest
from anyio import create_task_group, sleep
from pycrdt import (
    Awareness,
    Doc,
    Encoder,
    YMessageType,
    create_awareness_message,
    is_awareness_disconnect_message,
    read_message,
    write_message,
)

pytestmark = pytest.mark.anyio

TEST_USER = {"username": str(uuid4()), "name": "Test user"}
REMOTE_CLIENT_ID = 853790970
REMOTE_USER = {
    "user": {
        "username": "2460ab00fd28415b87e49ec5aa2d482d",
        "name": "Anonymous Ersa",
        "display_name": "Anonymous Ersa",
        "initials": "AE",
        "avatar_url": None,
        "color": "var(--jp-collaborator-color7)",
    }
}


def create_awareness_update(client_id, user, clock=1) -> bytes:
    if isinstance(user, str):
        new_user_str = user
    else:
        new_user_str = json.dumps(user, separators=(",", ":"))
    encoder = Encoder()
    encoder.write_var_uint(1)
    encoder.write_var_uint(client_id)
    encoder.write_var_uint(clock)
    encoder.write_var_string(new_user_str)
    return encoder.to_bytes()


def test_awareness_get_local_state():
    ydoc = Doc()
    awareness = Awareness(ydoc)

    assert awareness.get_local_state() == {}


def test_awareness_set_local_state():
    ydoc = Doc()
    awareness = Awareness(ydoc)
    changes = []

    def callback(topic, event):
        changes.append((topic, event))

    awareness.observe(callback)

    awareness.set_local_state({"foo": "bar"})
    assert awareness.get_local_state() == {"foo": "bar"}

    awareness.set_local_state(None)
    assert awareness.get_local_state() is None

    assert changes == [
        ("change", ({"added": [], "updated": [awareness.client_id], "removed": []}, "local")),
        ("update", ({"added": [], "updated": [awareness.client_id], "removed": []}, "local")),
        ("change", ({"added": [], "updated": [], "removed": [awareness.client_id]}, "local")),
        ("update", ({"added": [], "updated": [], "removed": [awareness.client_id]}, "local")),
    ]


def test_awareness_set_local_state_field():
    ydoc = Doc()
    awareness = Awareness(ydoc)

    awareness.set_local_state_field("new_field", "new_value")
    assert awareness.get_local_state() == {"new_field": "new_value"}


def test_awareness_add_user():
    ydoc = Doc()
    awareness = Awareness(ydoc)
    changes = []

    def callback(topic, event):
        changes.append((topic, event))

    awareness.observe(callback)
    awareness.apply_awareness_update(
        create_awareness_update(REMOTE_CLIENT_ID, REMOTE_USER),
        "custom_origin",
    )
    assert len(changes) == 2
    assert changes[0] == (
        "change",
        (
            {
                "added": [REMOTE_CLIENT_ID],
                "updated": [],
                "removed": [],
            },
            "custom_origin",
        ),
    )
    assert changes[1] == (
        "update",
        (
            {
                "added": [REMOTE_CLIENT_ID],
                "updated": [],
                "removed": [],
            },
            "custom_origin",
        ),
    )
    assert awareness.states == {
        REMOTE_CLIENT_ID: REMOTE_USER,
        ydoc.client_id: {},
    }


def test_awareness_update_user():
    ydoc = Doc()
    awareness = Awareness(ydoc)

    # Add a remote user.
    awareness.apply_awareness_update(
        create_awareness_update(REMOTE_CLIENT_ID, REMOTE_USER),
        "custom_origin",
    )

    # Update it.
    remote_user = deepcopy(REMOTE_USER)
    remote_user["user"]["name"] = "New user name"
    changes = []

    def callback(topic, event):
        changes.append((topic, event))

    awareness.observe(callback)
    awareness.apply_awareness_update(
        create_awareness_update(REMOTE_CLIENT_ID, remote_user, 2),
        "custom_origin",
    )

    assert len(changes) == 2
    assert changes[0] == (
        "change",
        (
            {
                "added": [],
                "updated": [REMOTE_CLIENT_ID],
                "removed": [],
            },
            "custom_origin",
        ),
    )
    assert changes[1] == (
        "update",
        (
            {
                "added": [],
                "updated": [REMOTE_CLIENT_ID],
                "removed": [],
            },
            "custom_origin",
        ),
    )
    assert awareness.states == {
        REMOTE_CLIENT_ID: remote_user,
        ydoc.client_id: {},
    }


def test_awareness_remove_user():
    ydoc = Doc()
    awareness = Awareness(ydoc)

    # Add a remote user.
    awareness.apply_awareness_update(
        create_awareness_update(REMOTE_CLIENT_ID, REMOTE_USER),
        "custom_origin",
    )

    # Remove it.
    changes = []

    def callback(topic, event):
        changes.append((topic, event))

    awareness.observe(callback)
    awareness.apply_awareness_update(
        create_awareness_update(REMOTE_CLIENT_ID, "null", 2),
        "custom_origin",
    )

    assert len(changes) == 2
    assert changes[0] == (
        "change",
        (
            {
                "added": [],
                "updated": [],
                "removed": [REMOTE_CLIENT_ID],
            },
            "custom_origin",
        ),
    )
    assert changes[1] == (
        "update",
        (
            {
                "added": [],
                "updated": [],
                "removed": [REMOTE_CLIENT_ID],
            },
            "custom_origin",
        ),
    )
    assert awareness.states == {ydoc.client_id: {}}


def test_awareness_increment_clock():
    ydoc = Doc()
    awareness = Awareness(ydoc)
    changes = []

    def callback(topic, event):
        changes.append((topic, event))

    awareness.observe(callback)
    awareness.apply_awareness_update(
        create_awareness_update(awareness.client_id, "null"),
        "custom_origin",
    )
    assert len(changes) == 2
    assert changes[0] == (
        "change",
        (
            {
                "added": [],
                "updated": [],
                "removed": [awareness.client_id],
            },
            "custom_origin",
        ),
    )
    assert changes[1] == (
        "update",
        (
            {
                "added": [],
                "updated": [],
                "removed": [awareness.client_id],
            },
            "custom_origin",
        ),
    )
    assert awareness.meta.get(awareness.client_id, {}).get("clock") == 2


def test_awareness_observes():
    ydoc = Doc()
    awareness = Awareness(ydoc)

    changes1 = []
    changes2 = []

    def callback1(topic, value):
        changes1.append((topic, value))

    def callback2(topic, value):
        changes2.append((topic, value))

    awareness.observe(callback1)
    sub2 = awareness.observe(callback2)
    awareness.apply_awareness_update(
        create_awareness_update(REMOTE_CLIENT_ID, REMOTE_USER),
        "custom_origin",
    )
    assert len(changes1) == 2
    assert len(changes2) == 2

    changes1.clear()
    changes2.clear()

    awareness.unobserve(sub2)
    awareness.apply_awareness_update(
        create_awareness_update(REMOTE_CLIENT_ID, "null"),
        "custom_origin",
    )
    assert len(changes1) == 2
    assert len(changes2) == 0


def test_awareness_observes_local_change():
    ydoc = Doc()
    awareness = Awareness(ydoc)
    changes = []

    def callback(topic, value):
        changes.append((topic, value))

    awareness.observe(callback)
    awareness.set_local_state_field("new_field", "new_value")
    assert len(changes) == 2
    assert changes[0] == (
        "change",
        (
            {
                "added": [],
                "removed": [],
                "updated": [ydoc.client_id],
            },
            "local",
        ),
    )
    assert changes[1] == (
        "update",
        (
            {
                "added": [],
                "removed": [],
                "updated": [ydoc.client_id],
            },
            "local",
        ),
    )


def test_awareness_encode():
    ydoc = Doc()
    awareness = Awareness(ydoc)
    changes = []

    def callback(topic, value):
        changes.append((topic, value))

    awareness.observe(callback)
    awareness.set_local_state_field("new_field", "new_value")
    awareness_update = awareness.encode_awareness_update(changes[0][1][0]["updated"])
    assert awareness_update == create_awareness_update(
        awareness.client_id, awareness.get_local_state()
    )
    message = create_awareness_message(awareness_update)
    assert message[0] == YMessageType.AWARENESS
    assert read_message(message[1:]) == awareness_update


def test_awareness_encode_wrong_id():
    ydoc = Doc()
    awareness = Awareness(ydoc)
    with pytest.raises(TypeError):
        awareness.encode_awareness_update([10])


async def test_awareness_periodic_updates():
    ydoc = Doc()
    outdated_timeout = 200
    awareness = Awareness(ydoc, outdated_timeout=outdated_timeout)
    remote_client_id = 0
    awareness._meta[remote_client_id] = {"clock": 0, "lastUpdated": 0}
    awareness._states[remote_client_id] = {}
    changes = []

    def callback(topic, value):
        changes.append((topic, value))

    awareness.observe(callback)
    async with create_task_group() as tg:
        await tg.start(awareness.start)
        with pytest.raises(RuntimeError) as excinfo:
            await tg.start(awareness.start)
        assert str(excinfo.value) == "Awareness already started"
        await sleep((outdated_timeout - outdated_timeout / 10) / 1000)
        awareness.remove_awareness_states([awareness.client_id], "local")
        await sleep(outdated_timeout / 1000)
        await awareness.stop()
        with pytest.raises(RuntimeError) as excinfo:
            await awareness.stop()
        assert str(excinfo.value) == "Awareness not started"

    assert len(changes) == 5
    assert changes[0] == (
        "change",
        (
            {
                "added": [],
                "removed": [remote_client_id],
                "updated": [],
            },
            "timeout",
        ),
    )
    assert changes[1] == (
        "update",
        (
            {
                "added": [],
                "removed": [remote_client_id],
                "updated": [],
            },
            "timeout",
        ),
    )
    assert changes[2] == (
        "update",
        (
            {
                "added": [],
                "removed": [],
                "updated": [awareness.client_id],
            },
            "local",
        ),
    )
    assert changes[3] == (
        "change",
        (
            {
                "added": [],
                "removed": [awareness.client_id],
                "updated": [],
            },
            "local",
        ),
    )
    assert changes[4] == (
        "update",
        (
            {
                "added": [],
                "removed": [awareness.client_id],
                "updated": [],
            },
            "local",
        ),
    )


def test_awareness_disconnection():
    # Should return True if it is a disconnection message
    update = write_message(create_awareness_update(REMOTE_CLIENT_ID, "null"))
    assert is_awareness_disconnect_message(update)

    # Should return False if it is not a disconnection message
    update = write_message(create_awareness_update(REMOTE_CLIENT_ID, "{}"))
    assert not is_awareness_disconnect_message(update)



import pytest
from pycrdt import Doc, Snapshot, Text


def test_snapshot_encode_roundtrip():
    doc = Doc(client_id=1, skip_gc=True)
    doc["test"] = Text()
    with doc.transaction():
        text = doc["test"]
        text.insert(0, "hello world")
    # Encode snapshot
    snapshot = Snapshot.from_doc(doc)
    encoded = snapshot.encode()
    # Decode snapshot
    snapshot2 = Snapshot.decode(encoded)
    encoded2 = snapshot2.encode()
    assert encoded == encoded2

    # Should fail on invalid bytes
    with pytest.raises(ValueError):
        Snapshot.decode(b"notavalidsnapshot")


def test_snapshot_multiple_edits():
    doc = Doc(client_id=2, skip_gc=True)
    doc["test"] = Text()
    with doc.transaction():
        text = doc["test"]
        text.insert(0, "abc")
    with doc.transaction():
        text = doc["test"]
        text.insert(3, "def")
    snapshot = Snapshot.from_doc(doc)
    encoded = snapshot.encode()
    assert isinstance(encoded, bytes)
    assert len(encoded) > 0
    # The encoded snapshot should change after edits
    with doc.transaction():
        text = doc["test"]
        text.insert(6, "ghi")
    snapshot2 = Snapshot.from_doc(doc)
    encoded2 = snapshot2.encode()
    assert encoded != encoded2


def test_doc_from_snapshot():
    doc = Doc(client_id=42, skip_gc=True)
    doc["test"] = Text()
    with doc.transaction():
        text = doc["test"]
        text.insert(0, "old")
    snapshot = Snapshot.from_doc(doc)
    # Make additional changes after snapshot
    with doc.transaction():
        text = doc["test"]
        text.insert(0, "new ")
    # Restore to snapshot in a new doc
    doc_restored = Doc.from_snapshot(snapshot, doc)
    # Should revert to snapshot state in restored doc
    assert "test" in doc_restored
    with doc_restored.transaction():
        text = doc_restored["test"]
    assert str(text) == "old"
    # Original doc should have both changes
    with doc.transaction():
        text = doc["test"]
    assert str(text) == "new old"



from typing import TypedDict, cast

import pytest
from pycrdt import Array, Doc, Map, Text, TypedArray, TypedMap, TypedDoc


@pytest.mark.mypy_testing
def mypy_test_array():
    doc = Doc[Array[int]]()
    array0 = doc.get("array0", type=Array)
    array0.append(0)
    array0.append("foo")  # E: Argument 1 to "append" of "Array" has incompatible type "str"; expected "int"  [arg-type]
    doc.get("array1", type=Array[str])  # E: Argument "type" to "get" of "Doc" has incompatible type "type[pycrdt._array.Array[str]]"; expected "type[pycrdt._array.Array[int]]"


@pytest.mark.mypy_testing
def mypy_test_uniform_map():
    doc = Doc[Map[bool]]()
    map0 = doc.get("map0", type=Map)
    map0["foo"] = True
    map0["foo"] = 3  # E: Incompatible types in assignment (expression has type "int", target has type "bool")
    v0: str = map0.pop("foo")  # E: Incompatible types in assignment (expression has type "bool", variable has type "str")
    v1: bool = map0.pop("foo")


@pytest.mark.mypy_testing
def mypy_test_typed_map():
    doc = Doc[Map]()

    MyMap = TypedDict(
        "MyMap",
        {
            "name": str,
            "toggle": bool,
            "nested": Array[bool],
        },
    )
    map0 = cast(MyMap, doc.get("map0", type=Map))
    map0["name"] = "foo"
    map0["toggle"] = False
    map0["toggle"] = 3  # E: Value of "toggle" has incompatible type "int"; expected "bool"
    array0 = Array([1, 2, 3])
    map0["nested"] = array0  # E: Value of "nested" has incompatible type "Array[int]"; expected "Array[bool]"
    array1 = Array([False, True])
    map0["nested"] = array1
    v0: str = map0["name"]
    v1: str = map0["toggle"]  # E: Incompatible types in assignment (expression has type "bool", variable has type "str")
    v2: bool = map0["toggle"]
    map0["key0"]  # E: TypedDict "MyMap@30" has no key "key0"


@pytest.mark.mypy_testing
def mypy_test_uniform_doc():
    doc = Doc[Text]()
    doc.get("text0", type=Text)
    doc.get("array0", type=Array)  # E: Argument "type" to "get" of "Doc" has incompatible type "type[pycrdt._array.Array[Any]]"; expected "type[Text]"
    doc.get("Map0", type=Map)  # E:  Argument "type" to "get" of "Doc" has incompatible type "type[pycrdt._map.Map[Any]]"; expected "type[Text]"


@pytest.mark.mypy_testing
def mypy_test_typed_doc():
    MyMap = TypedDict(
        "MyMap",
        {
            "name": str,
            "toggle": bool,
            "nested": Array[bool],
        },
    )

    MyDoc = TypedDict(
        "MyDoc",
        {
            "text0": Text,
            "array0": Array[int],
            "map0": MyMap,
        }
    )
    doc = cast(MyDoc, Doc())
    map0 = cast(MyMap, Map())
    doc["map0"] = map0
    doc["map0"] = Array()  # E: Value of "map0" has incompatible type "Array[Never]"; expected "MyMap@62"
    doc["map0"] = Map()  # E: Value of "map0" has incompatible type "Map[Never]"; expected "MyMap@62"
    doc["text0"] = Text()
    doc["array0"] = Array[bool]()  # E: Value of "array0" has incompatible type "Array[bool]"; expected "Array[int]"
    doc["array0"] = Array()


@pytest.mark.mypy_testing
def mypy_test_typed():

    class MyTypedArray(TypedArray[bool]):
        type: bool

    class MyTypedMap0(TypedMap):
        k0: bool

    class MyTypedMap1(TypedMap):
        key0: str
        key1: int
        key2: MyTypedMap0

    class MySubTypedDoc(TypedDoc):
        my_typed_map: MyTypedMap1

    class MyTypedDoc(MySubTypedDoc):
        my_array: MyTypedArray

    my_typed_doc = MyTypedDoc()
    my_typed_doc.my_typed_map.key0 = "foo"
    my_typed_doc.my_typed_map.key0 = 3  # E: Incompatible types in assignment (expression has type "int", variable has type "str")
    my_typed_doc.my_typed_map.key1 = 123
    my_typed_doc.my_typed_map.key1 = "bar"  # E: Incompatible types in assignment (expression has type "str", variable has type "int")
    my_typed_doc.my_typed_map.key2.k0 = 3  # E: Incompatible types in assignment (expression has type "int", variable has type "bool")
    my_typed_doc.my_typed_map.key2.k0 = False
    my_typed_doc.my_typed_map.key2.k1  # E: "MyTypedMap0" has no attribute "k1"
    my_typed_doc.my_array.append(True)
    my_typed_doc.my_array.append(2)  # E: Argument 1 to "append" of "TypedArray" has incompatible type "int"; expected "bool"
    my_typed_doc.my_wrong_root  # E: "MyTypedDoc" has no attribute "my_wrong_root"
    my_typed_doc.my_typed_map.wrong_key  # E: "MyTypedMap1" has no attribute "wrong_key"



import pytest
from anyio import TASK_STATUS_IGNORED, Event, create_task_group
from anyio.abc import TaskStatus
from pycrdt import Array, Assoc, Doc, Map, StickyIndex, Text

pytestmark = pytest.mark.anyio

hello = "Hello"
world = ", World"
sir = " Sir"
punct = "!"


def test_iterate():
    doc = Doc()
    doc["text"] = text = Text("abc")
    assert [char for char in text] == ["a", "b", "c"]


def test_str():
    doc1 = Doc()
    text1 = Text()
    doc1["text"] = text1
    with doc1.transaction():
        text1 += hello
        with doc1.transaction():
            text1 += world
        text1 += punct

    assert str(text1) == hello + world + punct

    doc2 = Doc()
    array2 = Array()
    doc2["array"] = array2
    text2 = Text("val")
    map2 = Map({"key": text2})
    array2.append(map2)
    assert str(array2) == '[{"key":"val"}]'


def test_api():
    doc = Doc()
    text = Text(hello + punct)

    with pytest.raises(RuntimeError) as excinfo:
        text.integrated
    assert str(excinfo.value) == "Not integrated in a document yet"

    with pytest.raises(RuntimeError) as excinfo:
        text.doc
    assert str(excinfo.value) == "Not integrated in a document yet"

    assert text.is_prelim
    assert text.prelim == hello + punct
    assert not text.is_integrated

    doc["text"] = text
    assert str(text) == hello + punct
    text.insert(len(hello), world)
    assert str(text) == hello + world + punct
    text.clear()
    assert len(text) == 0
    text[:] = hello + world + punct
    assert str(text) == hello + world + punct
    text[len(hello) : len(hello) + len(world)] = sir
    assert str(text) == hello + sir + punct
    # single character replacement
    text[len(text) - 1] = "?"
    assert str(text) == hello + sir + "?"
    # deletion with only an index
    del text[len(text) - 1]
    assert str(text) == hello + sir
    # deletion of an arbitrary range
    del text[len(hello) : len(hello) + len(sir)]
    assert str(text) == hello
    # deletion with start index == range length
    text += str(text)
    del text[len(hello) : 2 * len(hello)]
    assert str(text) == hello
    # deletion with a range of 0
    del text[len(hello) : len(hello)]
    assert str(text) == hello
    assert "".join([char for char in text]) == hello
    assert "el" in text

    with pytest.raises(RuntimeError) as excinfo:
        del text["a"]
    assert str(excinfo.value) == "Index not supported: a"

    with pytest.raises(RuntimeError) as excinfo:
        text["a"] = "b"
    assert str(excinfo.value) == "Index not supported: a"

    with pytest.raises(RuntimeError) as excinfo:
        text[1] = "ab"
    assert str(excinfo.value) == "Single item assigned value must have a length of 1, not 2"


def test_to_py():
    doc = Doc()
    doc["text"] = text = Text(hello)
    assert text.to_py() == hello


def test_prelim():
    text = Text(hello)
    assert text.to_py() == hello


def test_slice():
    doc = Doc()
    doc["text"] = text = Text(hello)

    for i, c in enumerate(hello):
        assert text[i] == c

    with pytest.raises(RuntimeError) as excinfo:
        text[1::2] = "a"
    assert str(excinfo.value) == "Step not supported"

    with pytest.raises(RuntimeError) as excinfo:
        text[-1:] = "a"
    assert str(excinfo.value) == "Negative start not supported"

    with pytest.raises(RuntimeError) as excinfo:
        text[:-1] = "a"
    assert str(excinfo.value) == "Negative stop not supported"


def test_formatting():
    doc = Doc()
    doc["text"] = text = Text("")

    text.insert(0, "hello ")
    assert len(text) == len("hello "), str(text)
    text.insert(len(text), "world", {"bold": True})
    text.insert(len(text), "! I have formatting!", {})
    text.format(len("hello world! "), len("hello world! I have formatting!") + 1, {"font-size": 32})
    text.insert_embed(len(text), b"png blob", {"type": "image"})

    diff = text.diff()

    assert diff == [
        ("hello ", None),
        ("world", {"bold": True}),
        ("! ", None),
        ("I have formatting!", {"font-size": 32}),
        (bytearray(b"png blob"), {"type": "image"}),
    ]


def test_observe():
    doc = Doc()
    doc["text"] = text = Text()
    events = []

    def callback(event):
        nonlocal text
        with pytest.raises(RuntimeError) as excinfo:
            text += world
        assert (
            str(excinfo.value)
            == "Read-only transaction cannot be used to modify document structure"
        )
        events.append(event)

    sub = text.observe(callback)  # noqa: F841
    text += hello
    assert str(events[0]) == """{target: Hello, delta: [{'insert': 'Hello'}], path: []}"""


async def test_iterate_events():
    doc = Doc()
    text = doc.get("text", type=Text)
    deltas = []

    async def iterate_events(done_event, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED):
        async with text.events() as events:
            task_status.started()
            idx = 0
            async for event in events:
                deltas.append(event.delta)
                if idx == 1:
                    done_event.set()
                    return
                idx += 1

    async with create_task_group() as tg:
        done_event = Event()
        await tg.start(iterate_events, done_event)
        text += "Hello"
        text += ", World!"
        await done_event.wait()
        text += " Goodbye."

    assert len(deltas) == 2
    assert deltas[0] == [{"insert": "Hello"}]
    assert deltas[1] == [{"retain": 5}, {"insert": ", World!"}]


@pytest.mark.parametrize("serialize", ["to_json", "encode"])
def test_sticky_index(serialize: str):
    first = "$$$"
    second = "-----*--"
    idx = second.index("*")

    doc0 = Doc()
    text0 = doc0.get("text", type=Text)
    text0 += first

    doc1 = Doc()
    text1 = doc1.get("text", type=Text)
    text1 += second

    assert text1[idx] == "*"
    sticky_index = text1.sticky_index(idx, Assoc.AFTER)
    assert sticky_index.assoc == Assoc.AFTER
    if serialize == "to_json":
        data = sticky_index.to_json()
        sticky_index = StickyIndex.from_json(data, text1)
    else:
        data = sticky_index.encode()
        sticky_index = StickyIndex.decode(data, text1)

    doc1.apply_update(doc0.get_update())
    assert str(text1) in (first + second, second + first)
    new_idx = sticky_index.get_index()
    assert text1[new_idx] == "*"


def test_sticky_index_transaction():
    doc = Doc()
    text = doc.get("text", type=Text)
    sticky_index = text.sticky_index(0, Assoc.BEFORE)
    data = sticky_index.to_json()
    sticky_index = StickyIndex.from_json(data)

    with pytest.raises(RuntimeError) as excinfo:
        sticky_index.get_index()

    assert str(excinfo.value) == "No transaction available"

    with doc.transaction() as txn:
        idx = sticky_index.get_index(txn)

    assert idx == 0



from contextlib import AsyncExitStack

import pytest
from anyio import create_memory_object_stream, create_task_group, sleep
from anyio.streams.stapled import StapledObjectStream
from pycrdt import (
    Channel,
    Doc,
    Provider,
    Text,
    YMessageType,
    create_sync_message,
    create_update_message,
    handle_sync_message,
)
from utils import StartStopContextManager

pytestmark = pytest.mark.anyio


class MyChannel(Channel):
    def __init__(self):
        self.doc = Doc()
        self.doc.observe(self._on_doc_change)
        self._stream_to_remote = StapledObjectStream(
            *create_memory_object_stream[bytes](float("inf"))
        )
        self._stream_from_remote = StapledObjectStream(
            *create_memory_object_stream[bytes](float("inf"))
        )

    def _on_doc_change(self, event):
        message = create_update_message(event.update)
        self._task_group.start_soon(self._stream_from_remote.send, message)

    async def __aenter__(self) -> "MyChannel":
        async with AsyncExitStack() as exit_stack:
            self._task_group = await exit_stack.enter_async_context(create_task_group())
            self._task_group.start_soon(self._run)
            self._exit_stack = exit_stack.pop_all()
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        await self._stream_from_remote.aclose()
        await self._stream_to_remote.aclose()
        return await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)

    async def _run(self):
        message = create_sync_message(self.doc)
        await self._stream_from_remote.send(message)
        async for message in self._stream_to_remote:
            message_type = message[0]
            if message_type == YMessageType.SYNC:
                reply = handle_sync_message(message[1:], self.doc)
                if reply is not None:
                    await self._stream_from_remote.send(reply)

    @property
    def path(self) -> str:
        return ""

    async def send(self, message: bytes) -> None:
        await self._stream_to_remote.send(message)

    async def recv(self) -> bytes:
        return await self._stream_from_remote.receive()


async def test_provider(service_api):
    local_doc = Doc()
    channel = MyChannel()
    provider = Provider(local_doc, channel)
    if service_api == "start_stop":
        provider = StartStopContextManager(provider)

    async with channel, provider:
        local_text = local_doc.get("text", type=Text)
        local_text += "Hello"
        await sleep(0.1)
        remote_doc = channel.doc
        remote_text = remote_doc.get("text", type=Text)
        assert str(remote_text) == "Hello"
        remote_text += ", World!"
        await sleep(0.1)
        assert str(local_text) == "Hello, World!"



import pytest
from pycrdt import Array, Doc, Map, Text, XmlElement, XmlFragment, XmlText


def test_plain_text():
    doc1 = Doc()
    frag = XmlFragment()
    doc1["test"] = frag
    with doc1.transaction():
        frag.children.append("Hello")
        with doc1.transaction():
            frag.children.append(", World")
        frag.children.append("!")

    assert str(frag) == "Hello, World!"


def test_api():
    doc = Doc()
    frag = XmlFragment(
        [
            XmlText("Hello "),
            XmlElement("em", {"class": "bold"}, [XmlText("World")]),
            XmlText("!"),
        ]
    )

    with pytest.raises(RuntimeError) as excinfo:
        frag.integrated
    assert str(excinfo.value) == "Not integrated in a document yet"

    with pytest.raises(RuntimeError) as excinfo:
        frag.doc
    assert str(excinfo.value) == "Not integrated in a document yet"

    with pytest.raises(ValueError):
        frag.to_py()

    doc["test"] = frag
    assert frag.parent is None
    assert str(frag) == 'Hello <em class="bold">World</em>!'
    assert len(frag.children) == 3
    assert str(frag.children[0]) == "Hello "
    assert str(frag.children[1]) == '<em class="bold">World</em>'
    assert str(frag.children[2]) == "!"
    assert list(frag.children) == [frag.children[0], frag.children[1], frag.children[2]]
    assert frag.children[0].parent == frag
    assert hash(frag.children[0].parent) == hash(frag)
    assert frag != object()

    frag.children.insert(1, XmlElement("strong", None, ["wonderful"]))
    frag.children.insert(2, " ")
    assert str(frag) == 'Hello <strong>wonderful</strong> <em class="bold">World</em>!'
    assert len(frag.children) == 5

    el = frag.children[3]
    assert el.tag == "em"
    assert len(el.attributes) == 1
    assert el.attributes.get("class") == "bold"
    assert el.attributes["class"] == "bold"
    assert "class" in el.attributes
    assert el.attributes.get("non-existent") is None
    assert "non-existent" not in el.attributes
    with pytest.raises(KeyError):
        el.attributes["non-existent"]
    assert list(el.attributes) == [("class", "bold")]

    del frag.children[2]
    del frag.children[1]
    assert str(frag) == 'Hello <em class="bold">World</em>!'


def test_text():
    text = XmlText("Hello")
    assert text.to_py() == "Hello"

    doc = Doc()

    with pytest.raises(ValueError):
        doc["test"] = XmlText("test")

    doc["test"] = XmlFragment([text])

    assert str(text) == "Hello"
    assert text.to_py() == "Hello"
    assert len(text) == len("Hello")

    text.clear()
    assert str(text) == ""

    text += "Goodbye"
    assert str(text) == "Goodbye"

    text.insert(1, " ")
    assert str(text) == "G oodbye"
    del text[1]
    assert str(text) == "Goodbye"

    text.insert(1, "  ")
    del text[1:3]
    assert str(text) == "Goodbye"

    assert text.diff() == [("Goodbye", None)]
    text.format(1, 3, {"bold": True})
    assert text.diff() == [
        ("G", None),
        ("oo", {"bold": True}),
        ("dbye", None),
    ]

    text.insert_embed(0, b"PNG!", {"type": "image"})
    assert text.diff() == [
        (b"PNG!", {"type": "image"}),
        ("G", None),
        ("oo", {"bold": True}),
        ("dbye", None),
    ]

    text.insert(len(text), " World!", {"href": "some-url"})
    assert text.diff() == [
        (b"PNG!", {"type": "image"}),
        ("G", None),
        ("oo", {"bold": True}),
        ("dbye", None),
        (" World!", {"href": "some-url"}),
    ]

    del text[0]
    assert text.diff() == [
        ("G", None),
        ("oo", {"bold": True}),
        ("dbye", None),
        (" World!", {"href": "some-url"}),
    ]

    del text[0:3]
    assert text.diff() == [
        ("dbye", None),
        (" World!", {"href": "some-url"}),
    ]

    array0 = Array(["foo", 2])
    text.insert_embed(0, array0)
    map0 = Map({"key": "val"})
    text.insert_embed(0, map0)
    text0 = Text("bar")
    text.insert_embed(0, text0)
    diff = text.diff()
    assert type(diff[0][0]) is type(text0.integrated)
    assert type(diff[1][0]) is type(map0.integrated)
    assert type(diff[2][0]) is type(array0.integrated)
    assert diff[3:] == [
        ("dbye", None),
        (" World!", {"href": "some-url"}),
    ]

    with pytest.raises(RuntimeError):
        del text[0:5:2]
    with pytest.raises(RuntimeError):
        del text[-1:5]
    with pytest.raises(RuntimeError):
        del text[1:-1]
    with pytest.raises(TypeError):
        del text["invalid"]

    doc["test2"] = XmlFragment([XmlText()])


def test_element_with_any_attribute():
    doc = Doc()

    doc["test"] = frag = XmlFragment()
    el = XmlElement("div")
    frag.children.append(el)
    el.attributes["class"] = {"a": True}
    assert el.attributes["class"] == {"a": True}
    assert list(el.attributes) == [("class", {"a": True})]
    assert len(el.attributes) == 1


def test_element():
    doc = Doc()

    with pytest.raises(ValueError):
        doc["test"] = XmlElement("test")

    with pytest.raises(ValueError):
        XmlElement()

    doc["test"] = frag = XmlFragment()

    el = XmlElement("div", {"class": "test"})
    frag.children.append(el)
    assert str(el) == '<div class="test"></div>'

    el = XmlElement("div", [("class", "test")])
    frag.children.append(el)
    assert str(el) == '<div class="test"></div>'

    el = XmlElement("div", None, [XmlText("Test")])
    frag.children.append(el)
    assert str(el) == "<div>Test</div>"

    el = XmlElement("div")
    frag.children.append(el)
    assert str(el) == "<div></div>"

    with pytest.raises(ValueError):
        el.to_py()

    el.attributes["class"] = "test"
    assert str(el) == '<div class="test"></div>'
    assert "class" in el.attributes
    assert el.attributes["class"] == "test"
    assert el.attributes.get("class") == "test"
    assert len(el.attributes) == 1
    assert list(el.attributes) == [("class", "test")]

    del el.attributes["class"]
    assert str(el) == "<div></div>"
    assert "class" not in el.attributes
    assert el.attributes.get("class") is None
    assert len(el.attributes) == 0
    assert list(el.attributes) == []

    node = XmlText("Hello")
    el.children.append(node)
    assert str(el) == "<div>Hello</div>"
    assert len(el.children) == 1
    assert str(el.children[0]) == "Hello"
    assert list(el.children) == [node]

    el.children[0] = XmlText("Goodbye")
    assert str(el) == "<div>Goodbye</div>"

    del el.children[0]
    assert str(el) == "<div></div>"

    el.children.append(XmlElement("foo"))
    el.children.append(XmlElement("bar"))
    el.children.append(XmlElement("baz"))
    assert str(el) == "<div><foo></foo><bar></bar><baz></baz></div>"

    del el.children[0:2]
    assert str(el) == "<div><baz></baz></div>"

    with pytest.raises(TypeError):
        del el.children["invalid"]
    with pytest.raises(IndexError):
        el.children[1]

    text = XmlText("foo")
    el.children.insert(0, text)
    assert str(el) == "<div>foo<baz></baz></div>"

    el2 = XmlElement("bar")
    el.children.insert(1, el2)
    assert str(el) == "<div>foo<bar></bar><baz></baz></div>"

    with pytest.raises(IndexError):
        el.children.insert(10, "test")
    with pytest.raises(ValueError):
        el.children.append(text)
    with pytest.raises(ValueError):
        el.children.append(el2)
    with pytest.raises(TypeError):
        el.children.append(object())


def test_observe():
    doc = Doc()
    doc["test"] = fragment = XmlFragment(["Hello world!"])
    events = []

    def callback(event):
        nonlocal fragment
        with pytest.raises(RuntimeError) as excinfo:
            fragment.children.append("text")
        assert (
            str(excinfo.value)
            == "Read-only transaction cannot be used to modify document structure"
        )
        events.append(event)

    sub = fragment.observe_deep(callback)  # noqa: F841

    fragment.children.append(XmlElement("em", None, ["This is a test"]))
    assert len(events) == 1
    assert len(events[0]) == 1
    assert events[0][0].children_changed is True
    assert str(events[0][0].target) == "Hello world!<em>This is a test</em>"
    assert events[0][0].path == []
    assert len(events[0][0].delta) == 2
    assert events[0][0].delta[0]["retain"] == 1
    assert str(events[0][0].delta[1]["insert"][0]) == "<em>This is a test</em>"

    events.clear()
    fragment.children[0].format(1, 3, {"bold": True})

    assert len(events) == 1
    assert len(events[0]) == 1
    assert str(events[0][0].target) == "H<bold>el</bold>lo world!"
    assert events[0][0].delta[0] == {"retain": 1}
    assert events[0][0].delta[1] == {"retain": 2, "attributes": {"bold": True}}


def test_xml_in_array():
    doc = Doc()
    array = doc.get("testmap", type=Array)
    frag = XmlFragment()
    array.append(frag)
    frag.children.append("Test XML!")

    assert len(array) == 1
    assert str(array[0]) == "Test XML!"

    with pytest.raises(TypeError):
        array.append(XmlText())
    with pytest.raises(TypeError):
        array.append(XmlElement("a"))
    assert len(array) == 1


def test_xml_in_map():
    doc = Doc()
    map = doc.get("testmap", type=Map)
    frag = map["testxml"] = XmlFragment()
    frag.children.append("Test XML!")

    assert len(map) == 1
    assert "testxml" in map
    assert str(map["testxml"]) == "Test XML!"

    with pytest.raises(TypeError):
        map["testtext"] = XmlText()
    with pytest.raises(TypeError):
        map["testel"] = XmlElement("a")
    assert len(map) == 1



from __future__ import annotations

import sys

import pytest
from pycrdt import Array, Doc, Map, TypedArray, TypedDoc, TypedMap


def test_typed_init():
    doc0 = Doc()

    typed_doc0 = TypedDoc(doc0)
    assert typed_doc0._ is doc0

    typed_doc1 = TypedDoc(typed_doc0)
    assert typed_doc1._ is doc0

    array0 = doc0.get("array0", type=Array)
    map0 = doc0.get("map0", type=Map)

    typed_array0 = TypedArray(array0)
    assert typed_array0._ is array0

    typed_array1 = TypedArray(typed_array0)
    assert typed_array1._ is array0

    typed_map0 = TypedMap(map0)
    assert typed_map0._ is map0

    typed_map1 = TypedMap(typed_map0)
    assert typed_map1._ is map0


class MyTypedArray(TypedArray[bool]):
    type: bool


class MyTypedMap0(TypedMap):
    k0: bool


class MyTypedMap1(TypedMap):
    key0: str
    key1: int
    key2: MyTypedMap0
    key3: Array[int]
    key4: str | int


class MySubTypedDoc(TypedDoc):
    my_typed_map: MyTypedMap1


class MyTypedDoc(MySubTypedDoc):
    my_array: MyTypedArray


def test_typed():
    doc = Doc()
    assert MyTypedDoc(doc)._ is doc

    my_typed_doc = MyTypedDoc()
    my_typed_doc.my_typed_map.key0 = "foo"

    with pytest.raises(TypeError) as excinfo:
        my_typed_doc.my_typed_map.key0 = 3
    assert str(excinfo.value) == (
        "Incompatible types in assignment (expression has type "
        "\"<class 'str'>\", variable has type \"<class 'int'>\")"
    )

    my_typed_doc.my_typed_map.key1 = 123
    my_typed_doc.my_typed_map.key2 = MyTypedMap0()
    my_typed_doc.my_typed_map.key2.k0 = False
    my_typed_doc.my_typed_map.key3 = Array([1, 2, 3])
    my_typed_doc.my_typed_map.key4 = "bar"
    assert my_typed_doc.my_typed_map.key4 == "bar"

    with pytest.raises(AttributeError) as excinfo:
        my_typed_doc.my_typed_map.wrong_key = "foo"
    assert str(excinfo.value) == '"<class \'test_typed.MyTypedMap1\'>" has no attribute "wrong_key"'

    with pytest.raises(AttributeError) as excinfo:
        my_typed_doc.my_typed_map.wrong_key
    assert str(excinfo.value) == '"<class \'test_typed.MyTypedMap1\'>" has no attribute "wrong_key"'

    assert len(my_typed_doc.my_array) == 0
    my_typed_doc.my_array.append(True)
    assert len(my_typed_doc.my_array) == 1
    assert my_typed_doc.my_array[0] is True
    my_typed_doc.my_array[0] = False
    assert my_typed_doc.my_array[0] is False
    my_typed_doc.my_array.extend([True])

    update = my_typed_doc._.get_update()

    my_other_doc = Doc()
    my_other_doc.apply_update(update)
    my_map = my_other_doc.get("my_typed_map", type=Map)
    assert my_map.to_py() == {
        "key0": "foo",
        "key1": 123.0,
        "key2": {"k0": False},
        "key3": [1.0, 2.0, 3.0],
        "key4": "bar",
    }
    my_array = my_other_doc.get("my_array", type=Array[bool])
    assert my_array.to_py() == [False, True]
